#include <shade/platform/drivers/vga_text_mode.h>
#include <shade/platform/ports.h>
#include <strings.h>

int row = 0;
int col = 0;
char color = PRINT_COLOR_WHITE | PRINT_COLOR_BLACK << 4;

void clear_row(int row) {
    struct Char* buffer = (struct Char*) 0xb8000;

    struct Char empty = (struct Char) {
        character: ' ',
        color: color,
    };

    for (int x = 0; x < NUM_COLS; x++) {
        buffer[x + NUM_COLS * row] = empty;
    }
    set_cursor_pos(col, row);
}

void clear_all() {
    for (int i = 0; i < NUM_ROWS; i++) {
        clear_row(i);
    }
    set_cursor_pos(col, row);
}

void print_newline() {
    struct Char* buffer = (struct Char*) 0xb8000;

    col = 0;

    if (row < NUM_ROWS - 1) {
        row++;
        return;
    }

    for (int row = 1; row < NUM_ROWS; row++) {
        for (int col = 0; col < NUM_COLS; col++) {
            struct Char character = buffer[col + NUM_COLS * row];
            buffer[col + NUM_COLS * (row - 1)] = character;
        }
    }

    clear_row(NUM_COLS - 1);
    set_cursor_pos(col, row);
}

bool is_hexchar(char character) {
    switch (character) {
        case '0':
        case '1':
        case '2':
        case '3':
        case '4':
        case '5':
        case '6':
        case '7':
        case '8':
        case '9':
        case 'a':
        case 'b':
        case 'c':
        case 'd':
        case 'e':
        case 'f':
            return true;
            break;
        default:
            return false;
            break;
    }
}

bool is_digit(char ch) {
  return (ch >= '0') && (ch <= '9');
}

unsigned int _atoi(const char** str) {
  unsigned int i = 0U;
  while (is_digit(**str)) {
    i = i * 10U + (unsigned int)(*((*str)++) - '0');
  }
  return i;
}



void print_char(char character) {
    // ANSI parsing
    // 1. is the current char \x1b?
    if (character == '\x1b') {
        // are we already printing an ansi string?
        if (is_parsing_ansi_string) {
            // string is invalid now
            is_parsing_ansi_string = false;
            return;
        } else {
            // no! set flag and continue
            is_parsing_ansi_string = true;
            return;
        }
    }

    // 2. are we working on an ansi string right now?
    if (is_parsing_ansi_string) {
        if (!ansi_string_found_bracket && character != '[') {
            // string is invalid, stop
            is_parsing_ansi_string = false;
            ansi_string_found_bracket = false;
            working_num = 0;
            ansi_string_found_m = false;
            ansi_string_found_semi = false;
            ansi_string_found_number = false;
            _flag = 0;
            _color = 0;
            _print_char(character);
            return;
        } else {
            // found bracket
            ansi_string_found_bracket = true;
            return;
        }
        if (ansi_string_found_bracket && !(ansi_string_found_semi || ansi_string_found_m)) {
            // looking for flag OR color, idk yet
            if (ansi_string_found_number) {
                if (character == 'm') {
                    if (ansi_string_found_m) {
                        // invalid
                        is_parsing_ansi_string = false;
                        ansi_string_found_bracket = false;
                        working_num = 0;
                        ansi_string_found_m = false;
                        ansi_string_found_semi = false;
                        ansi_string_found_number = false;
                        _flag = 0;
                        _color = 0;
                        _print_char(character);
                        return;
                    }
                    ansi_string_found_m = true;
                    // string is done
                    _color = working_num;
                    ansi_color_code_t code;
                    code.flag = _flag;
                    code.color = color;
                    handle_ansi_code(code);
                    is_parsing_ansi_string = false;
                    ansi_string_found_bracket = false;
                    working_num = 0;
                    ansi_string_found_m = false;
                    ansi_string_found_semi = false;
                    ansi_string_found_number = false;
                    _flag = 0;
                    _color = 0;
                    return;
                } else if (character == ';') {
                    if (ansi_string_found_semi) {
                        // invalid
                        is_parsing_ansi_string = false;
                        ansi_string_found_bracket = false;
                        working_num = 0;
                        ansi_string_found_m = false;
                        ansi_string_found_semi = false;
                        ansi_string_found_number = false;
                        _flag = 0;
                        _color = 0;
                        _print_char(character);
                        return;
                    }
                    _flag = working_num;
                    working_num = 0;
                    ansi_string_found_number = false;
                    ansi_string_found_semi = true;
                    return;
                }
            }
            if (!is_digit(character)) {
                // not a digit, print out and discard existing string
                is_parsing_ansi_string = false;
                ansi_string_found_bracket = false;
                working_num = 0;
                ansi_string_found_m = false;
                ansi_string_found_semi = false;
                ansi_string_found_number = false;
                _print_char(character);
                return;
            } else {
                // is digit, add to working
                working_num *= 10;
                working_num += _atoi(character);
                ansi_string_found_number = true;
            }
        }
        if (ansi_string_found_bracket && ansi_string_found_semi && !ansi_string_found_m) {
            if (ansi_string_found_number) {
                if (character == 'm') {
                    if (ansi_string_found_m) {
                        // invalid
                        is_parsing_ansi_string = false;
                        ansi_string_found_bracket = false;
                        working_num = 0;
                        ansi_string_found_m = false;
                        ansi_string_found_semi = false;
                        ansi_string_found_number = false;
                        _flag = 0;
                        _color = 0;
                        _print_char(character);
                        return;
                    }
                    ansi_string_found_m = true;
                    // string is done
                    _color = working_num;
                    ansi_color_code_t code;
                    code.flag = _flag;
                    code.color = _color;
                    handle_ansi_code(code);
                    is_parsing_ansi_string = false;
                    ansi_string_found_bracket = false;
                    working_num = 0;
                    ansi_string_found_m = false;
                    ansi_string_found_semi = false;
                    ansi_string_found_number = false;
                    _flag = 0;
                    _color = 0;
                    return;
                }
                if (!_is_digit(character)) {
                    // not a digit, print out and discard existing string
                    is_parsing_ansi_string = false;
                    ansi_string_found_bracket = false;
                    working_num = 0;
                    ansi_string_found_m = false;
                    ansi_string_found_semi = false;
                    ansi_string_found_number = false;
                    _flag = 0;
                    _color = 0;
                    _print_char(character);
                    return;
                } else {
                    // is digit, add to working
                    working_num *= 10;
                    working_num += _atoi(character);
                    ansi_string_found_number = true;
                }
            }
        }
    }

    _print_char(character);
    return;
}

void _print_char(char character) {
    struct Char* buffer = (struct Char*) 0xb8000;

    if (character == '\n') {
        print_newline();
        return;
    }

    if (col > NUM_COLS) {
        print_newline();
    }

    struct Char cr = (struct Char) {
        character: character,
        color: color,
    };
    buffer[col + NUM_COLS * row] = cr;

    col++;
    set_cursor_pos(col, row);
}

void print_str(char* str) {
    for (int i = 0; 1; i++) {
        char character = (char) str[i];

        if (character == '\0') {
            return;
        }

        print_char(character);
    }
}

void print_set_color(char foreground, char background) {
    color = foreground + (background << 4);
}

void set_cursor_pos(int col, int row) {
   int offset = col + NUM_COLS * row;
   outb(REG_SCREEN_CTRL, 14);
   outb(REG_SCREEN_DATA, (unsigned char)(offset >> 8));
   outb(REG_SCREEN_CTRL, 15);
   outb(REG_SCREEN_DATA, (unsigned char)(offset & 0xff)); 
}

void kernel_msg_ok(char* msg) {
	char ok_p1[] = "[ ";
	char ok_p2[] = "OK ";
	char ok_p3[] = "] ";
	
	print_set_color(PRINT_COLOR_WHITE, PRINT_COLOR_BLACK);
	print_str(ok_p1);
	
	print_set_color(PRINT_COLOR_GREEN, PRINT_COLOR_BLACK);
	print_str(ok_p2);
	
	print_set_color(PRINT_COLOR_WHITE, PRINT_COLOR_BLACK);
	print_str(ok_p3);
	print_str(msg);
}


void handle_ansi_code(ansi_color_code_t code) {
    char s[50];
    itoa(code.flag, s);
    print_str("ansi code: [");
    print_str(s);
    print_str(";");
    itoa(code.color, s);
    print_str(s);
    print_str("m found\n");
}